@@ -138,6 +138,8 @@ type 'a result =
 external ml_usb_init : unit -> unit = "ml_usb_init"
 external ml_usb_exit : unit -> unit = "ml_usb_exit"
 external ml_usb_set_debug : int -> unit = "ml_usb_set_debug"
+external ml_usb_get_next_timeout : unit -> float = "ml_usb_get_next_timeout"
+external ml_usb_handle_events : unit -> unit = "ml_usb_handle_events"
 external ml_usb_get_device_list : unit -> device list = "ml_usb_get_device_list"
 external ml_usb_get_bus_number : device -> int = "ml_usb_get_bus_number"
 external ml_usb_get_device_address : device -> int = "ml_usb_get_device_address"
@@ -146,15 +148,9 @@ external ml_usb_open : device -> device_handle = "ml_usb_open"
 external ml_usb_open_device_with_vid_pid : int -> int -> device_handle option = "ml_usb_open_device_with_vid_pid"
 external ml_usb_close : device_handle -> unit = "ml_usb_close"
 external ml_usb_get_device : device_handle -> device = "ml_usb_get_device"
-external ml_usb_claim_interface : device_handle -> interface -> unit = "ml_usb_claim_interface"
-external ml_usb_release_interface : device_handle -> interface -> unit = "ml_usb_release_interface"
-external ml_usb_get_configuration : device_handle -> configuration = "ml_usb_get_configuration"
-external ml_usb_set_configuration : device_handle -> configuration -> unit = "ml_usb_set_configuration"
 external ml_usb_kernel_driver_active : device_handle -> interface -> bool = "ml_usb_kernel_driver_active"
 external ml_usb_detach_kernel_driver : device_handle -> interface -> unit = "ml_usb_detach_kernel_driver"
 external ml_usb_attach_kernel_driver : device_handle -> interface -> unit = "ml_usb_attach_kernel_driver"
-external ml_usb_handle_events : unit -> unit = "ml_usb_handle_events"
-external ml_usb_collect_sources : Unix.file_descr list -> Unix.file_descr list -> Unix.file_descr list * Unix.file_descr list * float option = "ml_usb_collect_sources"
 external ml_usb_bulk_recv : device_handle * endpoint * int * string * int * int * (int result -> unit) -> transfer = "ml_usb_bulk_recv"
 external ml_usb_bulk_send : device_handle * endpoint * int * string * int * int * (int result -> unit) -> transfer = "ml_usb_bulk_send"
 external ml_usb_interrupt_recv : device_handle * endpoint * int * string * int * int * (int result -> unit) -> transfer = "ml_usb_interrupt_recv"
@@ -163,20 +159,68 @@ external ml_usb_control_recv : device_handle * endpoint * int * string * int * i
 external ml_usb_control_send : device_handle * endpoint * int * string * int * int * (int result -> unit) * recipient * request_type * request * int * int -> transfer = "ml_usb_control_send"
 external ml_usb_iso_recv : device_handle * endpoint * int * string * int * int * (int result list result -> unit) * int * int list -> transfer = "ml_usb_iso_recv"
 external ml_usb_iso_send : device_handle * endpoint * int * string * int * int * (int result list result -> unit) * int * int list -> transfer = "ml_usb_iso_send"
-external ml_usb_reset_device : device_handle -> unit = "ml_usb_reset_device"
 external ml_usb_cancel_transfer : transfer -> unit = "ml_usb_cancel_transfer"
-external ml_usb_set_interface_alt_setting : device_handle -> interface -> int -> unit = "ml_usb_set_interface_alt_setting"
-external ml_usb_clear_halt : device_handle -> endpoint -> unit = "ml_usb_clear_halt"
+
+external ml_usb_claim_interface_job : device_handle -> interface -> [ `claim_interface ] job = "ml_usb_claim_interface_job"
+external ml_usb_claim_interface_result : [ `claim_interface ] job -> unit = "ml_usb_claim_interface_result"
+external ml_usb_claim_interface_free : [ `claim_interface ] job -> unit = "ml_usb_claim_interface_free"
+
+external ml_usb_release_interface_job : device_handle -> interface -> [ `release_interface ] job = "ml_usb_release_interface_job"
+external ml_usb_release_interface_result : [ `release_interface ] job -> unit = "ml_usb_release_interface_result"
+external ml_usb_release_interface_free : [ `release_interface ] job -> unit = "ml_usb_release_interface_free"
+
+external ml_usb_get_configuration_job : device_handle -> [ `get_configuration ] job = "ml_usb_get_configuration_job"
+external ml_usb_get_configuration_result : [ `get_configuration ] job -> configuration = "ml_usb_get_configuration_result"
+external ml_usb_get_configuration_free : [ `get_configuration ] job -> unit = "ml_usb_get_configuration_free"
+
+external ml_usb_set_configuration_job : device_handle -> configuration -> [ `set_configuration ] job = "ml_usb_set_configuration_job"
+external ml_usb_set_configuration_result : [ `set_configuration ] job -> unit = "ml_usb_set_configuration_result"
+external ml_usb_set_configuration_free : [ `set_configuration ] job -> unit = "ml_usb_set_configuration_free"
+
+external ml_usb_set_interface_alt_setting_job : device_handle -> interface -> int -> [ `set_interface_alt_setting ] job = "ml_usb_set_interface_alt_setting_job"
+external ml_usb_set_interface_alt_setting_result : [ `set_interface_alt_setting ] job -> unit = "ml_usb_set_interface_alt_setting_result"
+external ml_usb_set_interface_alt_setting_free : [ `set_interface_alt_setting ] job -> unit = "ml_usb_set_interface_alt_setting_free"
+
+external ml_usb_clear_halt_job : device_handle -> endpoint -> [ `clear_halt ] job = "ml_usb_clear_halt_job"
+external ml_usb_clear_halt_result : [ `clear_halt ] job -> unit = "ml_usb_clear_halt_result"
+external ml_usb_clear_halt_free : [ `clear_halt ] job -> unit = "ml_usb_clear_halt_free"
+
+external ml_usb_reset_device_job : device_handle -> [ `reset_device ] job = "ml_usb_reset_device_job"
+external ml_usb_reset_device_result : [ `reset_device ] job -> unit = "ml_usb_reset_device_result"
+external ml_usb_reset_device_free : [ `reset_device ] job -> unit = "ml_usb_reset_device_free"
 
 (* +-----------------------------------------------------------------+
-   | Event-loop integration                                          |
+   | Lwt integration                                                 |
    +-----------------------------------------------------------------+ *)
 
-let filter_select now select set_r set_w set_e timeout =
-  let set_r, set_w, timeout' = ml_usb_collect_sources set_r set_w in
-  let res = select set_r set_w set_e (Lwt_main.min_timeout timeout timeout') in
-  ml_usb_handle_events ();
-  res
+let timeout_fired = ref false
+let timeout_event = ref Lwt_engine.fake_event
+
+let enter_iter () =
+  timeout_fired := false;
+  let timeout = ml_usb_get_next_timeout () in
+  if timeout >= 0. then
+    timeout_event := Lwt_engine.on_timer timeout false (fun _ -> timeout_fired := true)
+
+let leave_iter () =
+  Lwt_engine.stop_event !timeout_event;
+  if !timeout_fired then ml_usb_handle_events ()
+
+let events = Hashtbl.create 42
+
+let insert_pollfd fd check_readable check_writable =
+  let acc = [] in
+  let acc = if check_readable then Lwt_engine.on_readable fd (fun _ -> ml_usb_handle_events ()) :: acc else acc in
+  let acc = if check_writable then Lwt_engine.on_writable fd (fun _ -> ml_usb_handle_events ()) :: acc else acc in
+  Hashtbl.add events fd acc
+
+let remove_pollfd fd =
+  List.iter Lwt_engine.stop_event (Hashtbl.find events fd);
+  Hashtbl.remove events fd
+
+let () =
+  Callback.register "ocaml-usb:insert-pollfd" insert_pollfd;
+  Callback.register "ocaml-usb:remove-pollfd" remove_pollfd
 
 (* +-----------------------------------------------------------------+
    | Initialization                                                  |
@@ -185,10 +229,13 @@ let filter_select now select set_r set_w set_e timeout =
 (* Every function of this module must take care of forcing this before
    doing anything: *)
 let init = lazy(
+  (* Initializes libusb. *)
   ml_usb_init ();
-  let hook = Lwt_sequence.add_l filter_select Lwt_main.select_filters in
-  let exit = lazy(Lwt_sequence.remove hook;
-                  ml_usb_exit ()) in
+  (* Integrate libusb timoeuts into lwt. *)
+  ignore (Lwt_sequence.add_r enter_iter Lwt_main.enter_iter_hooks);
+  ignore (Lwt_sequence.add_r leave_iter Lwt_main.leave_iter_hooks);
+  (* Cleanup libusb on exit. *)
+  let exit = lazy(ml_usb_exit ()) in
   at_exit (fun _ -> Lazy.force exit)
 )
 
@@ -219,13 +266,13 @@ let make_handle device_handle = {
 let check_handle handle =
   if handle.state = State_closed then failwith "device handle closed"
 
-let detach handle f =
+let detach handle job result free =
   Lwt_mutex.with_lock handle.mutex
     (fun () ->
        check_handle handle;
        handle.state <- State_detach;
        try_lwt
-         Lwt_preemptive.detach f ()
+         execute_job (job ()) result free
        finally
          if handle.state = State_detach then
            handle.state <- State_ok;
@@ -235,39 +282,75 @@ let get_bus_number = ml_usb_get_bus_number
 let get_device_address = ml_usb_get_device_address
 let get_max_packet_size ~device ~direction ~endpoint = ml_usb_get_max_packet_size device direction endpoint
 let open_device device = make_handle (ml_usb_open device)
+
 let close handle =
   if handle.state <> State_closed then begin
     handle.state <- State_closed;
     ml_usb_close handle.handle
   end
+
 let get_device handle = ml_usb_get_device handle.handle
+
 let claim_interface handle interface =
-  try_lwt
-    ml_usb_claim_interface handle.handle interface;
-    return ()
+  detach
+    handle
+    (fun () -> ml_usb_claim_interface_job handle.handle interface)
+    ml_usb_claim_interface_result
+    ml_usb_claim_interface_free
+
 let release_interface handle interface =
-  detach handle (fun () -> ml_usb_release_interface handle.handle interface)
+  detach
+    handle
+    (fun () -> ml_usb_release_interface_job handle.handle interface)
+    ml_usb_release_interface_result
+    ml_usb_release_interface_free
+
 let kernel_driver_active handle =
   check_handle handle;
   ml_usb_kernel_driver_active handle.handle
+
 let detach_kernel_driver handle =
   check_handle handle;
   ml_usb_detach_kernel_driver handle.handle
+
 let attach_kernel_driver handle =
   check_handle handle;
   ml_usb_attach_kernel_driver handle.handle
+
 let get_configuration handle =
-  try_lwt
-    check_handle handle;
-    return (ml_usb_get_configuration handle.handle)
-let set_configuration handle conf =
-  detach handle (fun () -> ml_usb_set_configuration handle.handle conf)
+  detach
+    handle
+    (fun () -> ml_usb_get_configuration_job handle.handle)
+    ml_usb_get_configuration_result
+    ml_usb_get_configuration_free
+
+let set_configuration handle configuration =
+  detach
+    handle
+    (fun () -> ml_usb_set_configuration_job handle.handle configuration)
+    ml_usb_set_configuration_result
+    ml_usb_set_configuration_free
+
 let set_interface_alt_setting handle interface alt_setting =
-  detach handle (fun () -> ml_usb_set_interface_alt_setting handle.handle interface alt_setting)
+  detach
+    handle
+    (fun () -> ml_usb_set_interface_alt_setting_job handle.handle interface alt_setting)
+    ml_usb_set_interface_alt_setting_result
+    ml_usb_set_interface_alt_setting_free
+
 let clear_halt handle endpoint =
-  detach handle (fun () -> ml_usb_clear_halt handle.handle endpoint)
+  detach
+    handle
+    (fun () -> ml_usb_clear_halt_job handle.handle endpoint)
+    ml_usb_clear_halt_result
+    ml_usb_clear_halt_free
+
 let reset_device handle =
-  detach handle (fun () -> ml_usb_reset_device handle.handle)
+  detach
+    handle
+    (fun () -> ml_usb_reset_device_job handle.handle)
+    ml_usb_reset_device_result
+    ml_usb_reset_device_free
 
 let open_device_with ~vendor_id ~product_id =
   Lazy.force init;
@@ -379,7 +462,7 @@ let handle_result func_name w = function
 
 let make_timeout = function
   | None -> 0
-  | Some t -> truncate (t *. 1000.0)
+  | Some t -> Pervasives.truncate (t *. 1000.0)
 
 let transfer name func ~handle ~endpoint ?timeout buffer offset length =
   check_handle handle;
@@ -397,7 +480,7 @@ let interrupt_send = transfer "interrupt_send" ml_usb_interrupt_send
 let control_transfer name func ~handle ~endpoint ?timeout ?(recipient=Device) ?(request_type=Standard) ~request ~value ~index buffer offset length =
   check_handle handle;
   if offset < 0 || length < 0 || offset > String.length buffer - length then invalid_arg ("USB." ^ name);
-  let waiter, wakener = Lwt.wait () in
+  let waiter, wakener = Lwt.task () in
   let transfer = func (handle.handle, endpoint, make_timeout timeout, buffer, offset, length, handle_result name wakener, recipient, request_type, request, value, index) in
   Lwt.on_cancel waiter (fun () -> ml_usb_cancel_transfer transfer);
   waiter
@@ -419,7 +502,7 @@ let iso_transfer name func ~handle ~endpoint ?timeout buffer offset lengths =
     List.iter (fun length -> if length < 0 then invalid_arg ("USB." ^ name)) lengths;
     let length = List.fold_left (+) 0 lengths in
     if offset < 0 || offset > String.length buffer - length then invalid_arg ("USB." ^ name);
-    let waiter, wakener = Lwt.wait () in
+    let waiter, wakener = Lwt.task () in
     let transfer = func (handle.handle, endpoint, make_timeout timeout, buffer, offset, length, handle_iso_result name wakener, List.length lengths, lengths) in
     Lwt.on_cancel waiter (fun () -> ml_usb_cancel_transfer transfer);
     waiter
@@ -468,7 +551,7 @@ let get_string_descriptor handle ?timeout ?lang_id ~index =
           ~index:0
           data 0 (String.length data) in
         if n < 4 then
-          fail (Failure "USB.get_string_descriptor: cannot retreive default lang id")
+          raise_lwt (Failure "USB.get_string_descriptor: cannot retreive default lang id")
         else
           return (Char.code data.[2] lor (Char.code data.[3] lsl 8))
   in
@@ -482,6 +565,6 @@ let get_string_descriptor handle ?timeout ?lang_id ~index =
     data 0 (String.length data) in
   let len = Char.code data.[0] in
   if Char.code data.[1] <> DT.string || len > n then
-    fail (Failure "USB.get_string_descriptor: invalid control packet")
+    raise_lwt (Failure "USB.get_string_descriptor: invalid control packet")
   else
     return (String.sub data 2 (len - 2))
